@@ -40,8 +40,6 @@ import jakarta.servlet.http.HttpSession;
 
 import org.apache.catalina.AccessLog;
 import org.apache.catalina.Globals;
-import org.apache.catalina.LifecycleException;
-import org.apache.catalina.LifecycleState;
 import org.apache.catalina.Session;
 import org.apache.catalina.connector.ClientAbortException;
 import org.apache.catalina.connector.Request;
@@ -557,7 +555,9 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
             this.pattern = pattern;
         }
         logElements = createLogElements();
-        cachedElements = createCachedElements(logElements);
+        if (logElements != null) {
+            cachedElements = createCachedElements(logElements);
+        }
     }
 
     /**
@@ -645,16 +645,6 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
 
     // --------------------------------------------------------- Public Methods
 
-    /**
-     * Log a message summarizing the specified request and response, according to the format specified by the
-     * <code>pattern</code> property.
-     *
-     * @param request  Request being processed
-     * @param response Response being processed
-     *
-     * @exception IOException      if an input/output error has occurred
-     * @exception ServletException if a servlet error has occurred
-     */
     @Override
     public void invoke(Request request, Response response) throws IOException, ServletException {
         if (tlsAttributeRequired) {
@@ -752,33 +742,6 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
     }
 
 
-    /**
-     * Start this component and implement the requirements of
-     * {@link org.apache.catalina.util.LifecycleBase#startInternal()}.
-     *
-     * @exception LifecycleException if this component detects a fatal error that prevents this component from being
-     *                                   used
-     */
-    @Override
-    protected synchronized void startInternal() throws LifecycleException {
-
-        setState(LifecycleState.STARTING);
-    }
-
-
-    /**
-     * Stop this component and implement the requirements of
-     * {@link org.apache.catalina.util.LifecycleBase#stopInternal()}.
-     *
-     * @exception LifecycleException if this component detects a fatal error that prevents this component from being
-     *                                   used
-     */
-    @Override
-    protected synchronized void stopInternal() throws LifecycleException {
-
-        setState(LifecycleState.STOPPING);
-    }
-
     /**
      * AccessLogElement writes the partial message into the buffer.
      */
@@ -1336,8 +1299,64 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
      * write time taken to process the request - %D, %T
      */
     protected static class ElapsedTimeElement implements AccessLogElement {
-        private final boolean micros;
-        private final boolean millis;
+        enum Style {
+            SECONDS {
+                @Override
+                public void append(CharArrayWriter buf, long time) {
+                    buf.append(Long.toString(TimeUnit.NANOSECONDS.toSeconds(time)));
+                }
+            },
+            SECONDS_FRACTIONAL {
+                @Override
+                public void append(CharArrayWriter buf, long time) {
+                    time = time / 1000000; // Convert to millis
+                    buf.append(Long.toString(time / 1000));
+                    buf.append('.');
+                    int remains = (int) (time % 1000);
+                    buf.append(Long.toString(remains / 100));
+                    remains = remains % 100;
+                    buf.append(Long.toString(remains / 10));
+                    buf.append(Long.toString(remains % 10));
+                }
+            },
+            MILLISECONDS {
+                @Override
+                public void append(CharArrayWriter buf, long time) {
+                    buf.append(Long.toString(TimeUnit.NANOSECONDS.toMillis(time)));
+                }
+            },
+            MICROSECONDS {
+                @Override
+                public void append(CharArrayWriter buf, long time) {
+                    buf.append(Long.toString(TimeUnit.NANOSECONDS.toMicros(time)));
+                }
+            },
+            NANOSECONDS {
+                @Override
+                public void append(CharArrayWriter buf, long time) {
+                    buf.append(Long.toString(time));
+                }
+            };
+
+            /**
+             * Append the time to the buffer in the appropriate format.
+             *
+             * @param buf The buffer to append to.
+             * @param time The time to log in nanoseconds.
+             */
+            public abstract void append(CharArrayWriter buf, long time);
+        }
+
+        private final Style style;
+
+        /**
+         * Creates a new ElapsedTimeElement that will log the time in the specified style.
+         *
+         * @param style The elapsed-time style to use.
+         */
+        public ElapsedTimeElement(Style style) {
+            this.style = style;
+        }
 
         /**
          * @param micros <code>true</code>, write time in microseconds - %D
@@ -1345,20 +1364,12 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
          *                   time in seconds - %T
          */
         public ElapsedTimeElement(boolean micros, boolean millis) {
-            this.micros = micros;
-            this.millis = millis;
+            this(micros ? Style.MICROSECONDS : millis ? Style.MILLISECONDS : Style.SECONDS);
         }
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            if (micros) {
-                buf.append(Long.toString(TimeUnit.NANOSECONDS.toMicros(time)));
-            } else if (millis) {
-                buf.append(Long.toString(TimeUnit.NANOSECONDS.toMillis(time)));
-            } else {
-                // second
-                buf.append(Long.toString(TimeUnit.NANOSECONDS.toSeconds(time)));
-            }
+            style.append(buf, time);
         }
     }
 
@@ -1508,12 +1519,15 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            StringBuilder value = new StringBuilder();
+            StringBuilder value = null;
             boolean first = true;
             Cookie[] cookies = request.getCookies();
             if (cookies != null) {
                 for (Cookie cookie : cookies) {
                     if (cookieNameToLog.equals(cookie.getName())) {
+                        if (value == null) {
+                            value = new StringBuilder();
+                        }
                         if (first) {
                             first = false;
                         } else {
@@ -1523,7 +1537,7 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
                     }
                 }
             }
-            if (value.length() == 0) {
+            if (value == null) {
                 buf.append('-');
             } else {
                 escapeAndAppend(value.toString(), buf);
@@ -1756,10 +1770,14 @@ public abstract class AbstractAccessLogValve extends ValveBase implements Access
                 return new DateAndTimeElement(name);
             case 'T':
                 // ms for milliseconds, us for microseconds, and s for seconds
-                if ("ms".equals(name)) {
-                    return new ElapsedTimeElement(false, true);
+                if ("ns".equals(name)) {
+                    return new ElapsedTimeElement(ElapsedTimeElement.Style.NANOSECONDS);
                 } else if ("us".equals(name)) {
-                    return new ElapsedTimeElement(true, false);
+                    return new ElapsedTimeElement(ElapsedTimeElement.Style.MICROSECONDS);
+                } else if ("ms".equals(name)) {
+                    return new ElapsedTimeElement(ElapsedTimeElement.Style.MILLISECONDS);
+                } else if ("fracsec".equals(name)) {
+                    return new ElapsedTimeElement(ElapsedTimeElement.Style.SECONDS_FRACTIONAL);
                 } else {
                     return new ElapsedTimeElement(false, false);
                 }
