@@ -29,8 +29,8 @@ import java.util.List;
 import java.util.Locale;
 import java.util.TimeZone;
 
-import javax.servlet.http.Cookie;
-import javax.servlet.http.HttpSession;
+import jakarta.servlet.http.Cookie;
+import jakarta.servlet.http.HttpSession;
 
 import org.apache.catalina.connector.Request;
 import org.apache.catalina.connector.Response;
@@ -41,9 +41,8 @@ import org.apache.juli.logging.LogFactory;
 import org.apache.tomcat.util.ExceptionUtils;
 
 /**
- * An implementation of the W3c Extended Log File Format. See
- * <a href="http://www.w3.org/TR/WD-logfile.html">WD-logfile-960323</a>
- * for more information about the format. The following fields are supported:
+ * An implementation of the W3c Extended Log File Format. See http://www.w3.org/TR/WD-logfile.html for more information
+ * about the format. The following fields are supported:
  * <ul>
  * <li><code>c-dns</code>: Client hostname (or ip address if <code>enableLookups</code> for the connector is false)</li>
  * <li><code>c-ip</code>: Client ip address</li>
@@ -55,17 +54,17 @@ import org.apache.tomcat.util.ExceptionUtils;
  * <li><code>date</code>: The date in yyyy-mm-dd format for GMT</li>
  * <li><code>s-dns</code>: The server dns entry</li>
  * <li><code>s-ip</code>: The server ip address</li>
- * <li><code>cs(xxx)</code>: The value of header xxx from client to server</li>
- * <li><code>sc(xxx)</code>: The value of header xxx from server to client</li>
+ * <li><code>cs(XXX)</code>: The value of header XXX from client to server</li>
+ * <li><code>sc(XXX)</code>: The value of header XXX from server to client</li>
  * <li><code>sc-status</code>: The status code</li>
  * <li><code>time</code>: Time the request was served</li>
  * <li><code>time-taken</code>: Time (in seconds) taken to serve the request</li>
  * <li><code>x-threadname</code>: Current request thread name (can compare later with stacktraces)</li>
- * <li><code>x-A(xxx)</code>: Pull xxx attribute from the servlet context</li>
- * <li><code>x-C(xxx)</code>: Pull the cookie(s) of the name xxx</li>
- * <li><code>x-O(xxx)</code>: Pull the all response header values xxx</li>
- * <li><code>x-R(xxx)</code>: Pull xxx attribute from the servlet request</li>
- * <li><code>x-S(xxx)</code>: Pull xxx attribute from the session</li>
+ * <li><code>x-A(XXX)</code>: Pull XXX attribute from the servlet context</li>
+ * <li><code>x-C(XXX)</code>: Pull the cookie(s) of the name XXX</li>
+ * <li><code>x-O(XXX)</code>: Pull the all response header values XXX</li>
+ * <li><code>x-R(XXX)</code>: Pull XXX attribute from the servlet request</li>
+ * <li><code>x-S(XXX)</code>: Pull XXX attribute from the session</li>
  * <li><code>x-P(...)</code>: Call request.getParameter(...) and URLencode it. Helpful to capture certain POST
  * parameters.</li>
  * <li>For any of the x-H(...) the following method will be called from the HttpServletRequest object</li>
@@ -81,6 +80,26 @@ import org.apache.tomcat.util.ExceptionUtils;
  * <li><code>x-H(scheme)</code>: getScheme</li>
  * <li><code>x-H(secure)</code>: isSecure</li>
  * </ul>
+ * <p>
+ * Log rotation can be on or off. This is dictated by the <code>rotatable</code> property.
+ * </p>
+ * <p>
+ * For UNIX users, another field called <code>checkExists</code> is also available. If set to true, the log file's
+ * existence will be checked before each logging. This way an external log rotator can move the file somewhere and
+ * Tomcat will start with a new file.
+ * </p>
+ * <p>
+ * For JMX junkies, a public method called <code>rotate</code> has been made available to allow you to tell this
+ * instance to move the existing log file to somewhere else and start writing a new log file.
+ * </p>
+ * <p>
+ * Conditional logging is also supported. This can be done with the <code>condition</code> property. If the value
+ * returned from ServletRequest.getAttribute(condition) yields a non-null value, the logging will be skipped.
+ * </p>
+ * <p>
+ * For extended attributes coming from a getAttribute() call, it is you responsibility to ensure there are no newline or
+ * control characters.
+ * </p>
  *
  * @author Peter Rossbach
  */
@@ -91,18 +110,19 @@ public class ExtendedAccessLogValve extends AccessLogValve {
     // -------------------------------------------------------- Private Methods
 
     /**
-     * Calls toString() on the object, wraps the result with double quotes (") and writes the result to the buffer. Any
-     * double quotes appearing in the value are escaped using two double quotes (""). If the value is null or if
-     * toString() fails, '-' will be written to the buffer.
+     * Wrap the incoming value with double quotes (") and escape any double quotes appearing in the value using two
+     * double quotes ("").
      *
      * @param value - The value to wrap
-     * @param buf the buffer to write to
+     *
+     * @return '-' if null. Otherwise, toString() will be called on the object and the value will be wrapped in quotes
+     *             and any quotes will be escaped with 2 sets of quotes.
      */
-    static void wrap(Object value, CharArrayWriter buf) {
+    static String wrap(Object value) {
         String svalue;
+        // Does the value contain a " ? If so must encode it
         if (value == null || "-".equals(value)) {
-            buf.append('-');
-            return;
+            return "-";
         }
 
         try {
@@ -110,16 +130,27 @@ public class ExtendedAccessLogValve extends AccessLogValve {
         } catch (Throwable e) {
             ExceptionUtils.handleThrowable(e);
             /* Log error */
-            buf.append('-');
-            return;
+            return "-";
         }
 
-        buf.append('\"');
-        if (!svalue.isEmpty()) {
-            // Does the value contain a " ? If so must encode it
-            escapeAndAppend(svalue, buf, true);
+        /* Wrap all values in double quotes. */
+        StringBuilder buffer = new StringBuilder(svalue.length() + 2);
+        buffer.append('\"');
+        int i = 0;
+        while (i < svalue.length()) {
+            int j = svalue.indexOf('\"', i);
+            if (j == -1) {
+                buffer.append(svalue.substring(i));
+                i = svalue.length();
+            } else {
+                buffer.append(svalue.substring(i, j + 1));
+                buffer.append('"');
+                i = j + 1;
+            }
         }
-        buf.append('\"');
+
+        buffer.append('\"');
+        return buffer.toString();
     }
 
     @Override
@@ -140,8 +171,8 @@ public class ExtendedAccessLogValve extends AccessLogValve {
         // Milli-seconds in 24 hours
         private static final long INTERVAL = (1000 * 60 * 60 * 24);
 
-        private static final ThreadLocal<ElementTimestampStruct> currentDate =
-                ThreadLocal.withInitial(() -> new ElementTimestampStruct("yyyy-MM-dd"));
+        private static final ThreadLocal<ElementTimestampStruct> currentDate = ThreadLocal
+                .withInitial(() -> new ElementTimestampStruct("yyyy-MM-dd"));
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
@@ -159,8 +190,8 @@ public class ExtendedAccessLogValve extends AccessLogValve {
         // Milli-seconds in a second
         private static final long INTERVAL = 1000;
 
-        private static final ThreadLocal<ElementTimestampStruct> currentTime =
-                ThreadLocal.withInitial(() -> new ElementTimestampStruct("HH:mm:ss"));
+        private static final ThreadLocal<ElementTimestampStruct> currentTime = ThreadLocal
+                .withInitial(() -> new ElementTimestampStruct("HH:mm:ss"));
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
@@ -183,7 +214,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            wrap(request.getHeader(header), buf);
+            buf.append(wrap(request.getHeader(header)));
         }
     }
 
@@ -196,7 +227,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            wrap(response.getHeader(header), buf);
+            buf.append(wrap(response.getHeader(header)));
         }
     }
 
@@ -209,7 +240,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            wrap(request.getContext().getServletContext().getAttribute(attribute), buf);
+            buf.append(wrap(request.getContext().getServletContext().getAttribute(attribute)));
         }
     }
 
@@ -238,7 +269,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             if (value.length() == 0) {
                 buf.append('-');
             } else {
-                wrap(value, buf);
+                buf.append(wrap(value.toString()));
             }
         }
     }
@@ -268,9 +299,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
                         }
                         buffer.append(iter.next());
                     }
-                    wrap(buffer, buf);
-                } else {
-                    buf.append('-');
+                    buf.append(wrap(buffer.toString()));
                 }
                 return;
             }
@@ -287,7 +316,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            wrap(request.getAttribute(attribute), buf);
+            buf.append(wrap(request.getAttribute(attribute)));
         }
     }
 
@@ -304,7 +333,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             if (request != null) {
                 session = request.getSession(false);
                 if (session != null) {
-                    wrap(session.getAttribute(attribute), buf);
+                    buf.append(wrap(session.getAttribute(attribute)));
                 }
             }
         }
@@ -329,13 +358,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
 
         @Override
         public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-            String parameterValue;
-            try {
-                parameterValue = request.getParameter(parameter);
-            } catch (IllegalStateException ise) {
-                parameterValue = null;
-            }
-            wrap(urlEncode(parameterValue), buf);
+            buf.append(wrap(urlEncode(request.getParameter(parameter))));
         }
     }
 
@@ -509,7 +532,19 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             if (tokenizer.hasSubToken()) {
                 String nextToken = tokenizer.getToken();
                 if ("taken".equals(nextToken)) {
-                    return new ElapsedTimeElement(false);
+                    nextToken = tokenizer.getToken();
+
+                    if ("ns".equals(nextToken)) {
+                        return new ElapsedTimeElement(ElapsedTimeElement.Style.NANOSECONDS);
+                    } else if("us".equals(nextToken)) {
+                        return new ElapsedTimeElement(ElapsedTimeElement.Style.MICROSECONDS);
+                    } else if ("ms".equals(nextToken)) {
+                        return new ElapsedTimeElement(ElapsedTimeElement.Style.MILLISECONDS);
+                    } else if ("fracsec".equals(nextToken)) {
+                        return new ElapsedTimeElement(ElapsedTimeElement.Style.SECONDS_FRACTIONAL);
+                    } else {
+                        return new ElapsedTimeElement(ElapsedTimeElement.Style.SECONDS);
+                    }
                 }
             } else {
                 return new TimeElement();
@@ -589,8 +624,10 @@ public class ExtendedAccessLogValve extends AccessLogValve {
                         public void addElement(CharArrayWriter buf, Date date, Request request, Response response,
                                 long time) {
                             String query = request.getQueryString();
-                            buf.append(request.getRequestURI());
-                            if (query != null) {
+                            if (query == null) {
+                                buf.append(request.getRequestURI());
+                            } else {
+                                buf.append(request.getRequestURI());
                                 buf.append('?');
                                 buf.append(request.getQueryString());
                             }
@@ -631,6 +668,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
     }
 
     protected AccessLogElement getProxyElement(PatternTokenizer tokenizer) throws IOException {
+        String token = null;
         if (tokenizer.hasSubToken()) {
             tokenizer.getToken();
             return new StringElement("-");
@@ -638,7 +676,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             tokenizer.getParameter();
             return new StringElement("-");
         }
-        log.error(sm.getString("extendedAccessLogValve.decodeError", tokenizer.getRemains()));
+        log.error(sm.getString("extendedAccessLogValve.decodeError", token));
         return null;
     }
 
@@ -685,63 +723,63 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getAuthType(), buf);
+                    buf.append(wrap(request.getAuthType()));
                 }
             };
         } else if ("remoteUser".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getRemoteUser(), buf);
+                    buf.append(wrap(request.getRemoteUser()));
                 }
             };
         } else if ("requestedSessionId".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getRequestedSessionId(), buf);
+                    buf.append(wrap(request.getRequestedSessionId()));
                 }
             };
         } else if ("requestedSessionIdFromCookie".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(String.valueOf(request.isRequestedSessionIdFromCookie()), buf);
+                    buf.append(wrap("" + request.isRequestedSessionIdFromCookie()));
                 }
             };
         } else if ("requestedSessionIdValid".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(String.valueOf(request.isRequestedSessionIdValid()), buf);
+                    buf.append(wrap("" + request.isRequestedSessionIdValid()));
                 }
             };
         } else if ("contentLength".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(String.valueOf(request.getContentLengthLong()), buf);
+                    buf.append(wrap("" + request.getContentLengthLong()));
                 }
             };
         } else if ("characterEncoding".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getCharacterEncoding(), buf);
+                    buf.append(wrap(request.getCharacterEncoding()));
                 }
             };
         } else if ("locale".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getLocale(), buf);
+                    buf.append(wrap(request.getLocale()));
                 }
             };
         } else if ("protocol".equals(parameter)) {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(request.getProtocol(), buf);
+                    buf.append(wrap(request.getProtocol()));
                 }
             };
         } else if ("scheme".equals(parameter)) {
@@ -755,7 +793,7 @@ public class ExtendedAccessLogValve extends AccessLogValve {
             return new AccessLogElement() {
                 @Override
                 public void addElement(CharArrayWriter buf, Date date, Request request, Response response, long time) {
-                    wrap(Boolean.valueOf(request.isSecure()), buf);
+                    buf.append(wrap("" + request.isSecure()));
                 }
             };
         }
